from abc import abstractmethod, ABCMeta

from rummy.player.player import Player
from rummy.ui.view import View


class PlayerController(metaclass=ABCMeta):

    @staticmethod
    @abstractmethod
    def show_start_turn(player: Player):
        pass

    @staticmethod
    @abstractmethod
    def show_end_turn(player: Player):
        pass

    @staticmethod
    @abstractmethod
    def show_discard(player: Player):
        pass

    @classmethod
    @abstractmethod
    def draw_card(cls, player):
        pass

    @classmethod
    @abstractmethod
    def discard_or_knock(cls, player):
        pass

    @staticmethod
    def show_knocked(player):
        if player.has_someone_knocked():
            View.render(View.prepare_template('/knocked.txt'))



from rummy.controller.player_controller import PlayerController
from rummy.player.player import Player
from rummy.ui.player_action_dialog import PlayerActionDialog
from rummy.ui.user_input import UserInput
from rummy.ui.view import View
from rummy.view.human_view import HumanView


class HumanController(PlayerController):

    @staticmethod
    def show_start_turn(player: Player):
        View.render(HumanView.turn_start(player))

    @staticmethod
    def show_end_turn(player: Player):
        View.render(HumanView.turn_end(player))

    @staticmethod
    def show_knocked(player):
        if player.has_someone_knocked():
            View.render(HumanView.knocked())

    @staticmethod
    def show_discard(player: Player):
        View.render(HumanView.discarded(player))

    @classmethod
    def draw_card(cls, player):
        if player.round.deck.has_discard():
            View.render(cls._choose_pick_up(player))
        else:
            player.take_from_deck()
            View.render('Drew from deck')

    @staticmethod
    def _choose_pick_up(player):
        user_input = UserInput.create_input(PlayerActionDialog.pick_up_or_draw())
        if user_input == 'p':
            player.take_from_discard()
            return 'Drawing from discard'
        else:
            player.take_from_deck()
            return 'Drawing from deck'

    @classmethod
    def discard_or_knock(cls, player):
        player.discard(cls._choose_discard(player))

    @staticmethod
    def _choose_discard(player):
        user_input = ''
        scores = player.melds.find_discard_scores(player.hand.get_hand())
        while user_input not in [str(i) for i in range(1, 9)]:
            if min(scores) <= 10 and not player.round.show_knocked:
                user_input = UserInput.create_input(PlayerActionDialog.choose_discard_or_knock())
                if user_input == "k":
                    player.knock()
                    continue
            else:
                user_input = UserInput.create_input(PlayerActionDialog.choose_discard())
        return user_input



# -*- coding: utf-8 -*-

from ansi_colours import AnsiColours as Colour

from rummy.view.round_view import RoundView


class Score:
    def __init__(self, players):
        self.players = players

    def get_current_game_scores(self):
        return ''.join(["%s: %s\n" % (p, p.get_game_score()) for p in self.players])

    def get_end_of_round_scores(self):
        output = ''
        for p in self.players:
            output += RoundView.end_of_round_scores(p)
        return output

    def update_player_scores(self):
        for p in self.players:
            p.update_score()

    def is_end_of_game(self):
        for p in self.players:
            if p.get_game_score() >= 100:
                return True
        return False

    def show_winners(self):
        winners = self.find_lowest_scores()
        if len(winners) == 1:
            return Colour.green("%s is the Winner!!" % winners[0])
        else:
            return Colour.green(", ".join([str(w) for w in winners]) + " are joint winners!")

    def find_lowest_scores(self):
        lowest = []
        for p in self.players:
            if not lowest:
                lowest = [p]
                continue
            if p.get_game_score() < lowest[0].get_game_score():
                lowest = [p]
            elif p.get_game_score() == lowest[0].get_game_score():
                lowest.append(p)
        return lowest



# coding=utf-8

from ansi_colours import AnsiColours as Colour

from rummy.ui.action_collection import ActionCollection
from rummy.ui.view import View


class UserInput:
    @staticmethod
    def create_input(action_collection: ActionCollection):
        user_input = None
        keys = []
        output = ''
        while True:
            for action in action_collection.actions:
                if isinstance(action.key, range):
                    output += "%s: %s\n" % (
                    str(action), '%s-%s' % (Colour.green(str(action.key[0])), Colour.green(str(action.key[-1]))))
                    keys.extend([str(key) for key in action.key])
                else:
                    output += "%s: %s\n" % (str(action), Colour.green(action.key))
                    keys.append(str(action.key))
            View.render(output)
            while user_input not in keys:
                user_input = UserInput.get_input('Select an option: ')
            break
        return user_input

    @staticmethod
    def get_input(*args, **kwargs):
        # Wrapper/abstraction layer for builtins.input
        player_input = input(*args, **kwargs)
        return player_input

    @staticmethod
    def get_pick_up_input():
        player_choice = ''
        while player_choice.lower() not in ['d', 'p']:
            player_choice = UserInput.get_input(
                "Enter " + Colour.green('d') + " to draw or " + Colour.green('p') + " to pickup discard: ")
        return player_choice



from random import choice

from rummy.controller.player_controller import PlayerController
from rummy.player.player import Player
from rummy.ui.view import View
from rummy.view.ai_view import AiView


class AiController(PlayerController):

    @staticmethod
    def show_start_turn(player: Player):
        output = ''
        if player.ai_only:
            output += AiView.turn_start(player)
        else:
            output += AiView.turn_start(player)
        output += AiView.thinking(player, 'Choosing pick up')
        View.render(output)

    @staticmethod
    def show_end_turn(player: Player):
        output = ''
        if player.ai_only:
            output += AiView.turn_end(player)
        output += AiView.thinking(player, 'Choosing card to discard')
        View.render(output)

    @staticmethod
    def show_knocked(player):
        if player.has_someone_knocked():
            View.render(AiView.knocked())

    @staticmethod
    def show_discard(player: Player):
        output = ''
        if player.ai_only:
            output += AiView.hand_data(player.hand.get_score())
        output += AiView.discarded(player.round.deck.inspect_discard())
        View.render(output)

    @classmethod
    def draw_card(cls, player):
        output = ''
        if player.round.deck.has_discard():
            current_score = player.hand.get_score()
            scores = player.melds.find_discard_scores(player.hand.get_hand(), player.round.deck.inspect_discard())
            if player.ai_only:
                output += AiView.discard_data(current_score, scores)
            output += cls._choose_pickup(player, current_score, scores)
        else:
            player.take_from_deck()
            output += AiView.thinking(player, 'Drawing from deck')
        View.render(output)

    @staticmethod
    def _choose_pickup(player, current_score, scores):
        output = ''
        if min(scores) < current_score - 4 or min(scores) <= 10:
            player.take_from_discard()
            output += AiView.thinking(player, 'Drawing from discard')
        else:
            player.take_from_deck()
            output += AiView.thinking(player, 'Drawing from deck')
        return output

    @classmethod
    def discard_or_knock(cls, player):
        scores = player.melds.find_discard_scores(player.hand.get_hand())
        score = min(scores)
        if score <= 10 and not player.round.knocked:
            player.round.knocked = True
        discard = cls._choose_discard(player, score, scores)
        player.round.deck.discard_card(discard)

    @staticmethod
    def _choose_discard(player, score, scores):
        if scores.count(score) > 1:
            choices = [(i, x) for (i, x) in enumerate(scores) if (x == score)]
            discard = choice(choices)[0]
        else:
            discard = scores.index(score)
        return player.hand.discard_card(discard)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rummy',
 'rummy.constants',
 'rummy.controller',
 'rummy.deck',
 'rummy.exception',
 'rummy.game',
 'rummy.player',
 'rummy.ui',
 'rummy.view']

package_data = \
{'': ['*'], 'rummy': ['templates/*']}

install_requires = \
['Pygments==2.2.0',
 'ansi-colours==0.2.6',
 'atomicwrites==1.2.1',
 'attrs==18.2.0',
 'bleach==3.0.2',
 'certifi==2018.10.15',
 'chardet==3.0.4',
 'codecov==2.0.15',
 'colorama==0.4.0',
 'coverage==4.5.1',
 'docutils==0.14',
 'idna==2.7',
 'more-itertools==4.3.0',
 'pathlib2==2.3.2',
 'pkginfo==1.4.2',
 'pluggy==0.8.0',
 'py==1.7.0',
 'pytest-cov==2.6.0',
 'pytest-mock==1.10.0',
 'pytest==3.10.0',
 'readme-renderer==24.0',
 'requests-toolbelt==0.8.0',
 'requests==2.20.1',
 'six==1.11.0',
 'text-template==0.1.4',
 'tqdm==4.28.1',
 'twine==1.12.1',
 'urllib3==1.24.1',
 'webencodings==0.5.1']

setup_kwargs = {
    'name': 'rummy',
    'version': '2.0.0',
    'description': 'Console Rummy game',
    'long_description': None,
    'author': 'Sarcoma',
    'author_email': 'sarcoma@live.co.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod

from text_template import TextTemplate

from rummy.constants.resource_path import TEMPLATE_PATH


class PlayerView:

    @staticmethod
    @abstractmethod
    def turn_start(player):
        pass

    @staticmethod
    @abstractmethod
    def turn_end(player):
        pass

    @staticmethod
    @abstractmethod
    def discarded(discard):
        pass

    @staticmethod
    def knocked():
        return TextTemplate.render(
            TEMPLATE_PATH + '/knocked.txt'
        )



from time import sleep

from text_template import TextTemplate

from rummy.constants.resource_path import TEMPLATE_PATH
from rummy.view.player_view import PlayerView


class AiView(PlayerView):

    @staticmethod
    def thinking(ai, action):
        output = TextTemplate.render(
            TEMPLATE_PATH + '/ai-thinking.txt', action=action)
        if not ai.ai_only:
            sleep(0.8)
        return output

    @staticmethod
    def turn_start(player):
        return TextTemplate.render(
            TEMPLATE_PATH + '/turn-start.txt',
            turn_number=player.round.turn,
            player_number=player.round.current_player + 1,
            discard=player.round.deck.show_discard()
        )

    @staticmethod
    def turn_end(player):
        return TextTemplate.render(
            TEMPLATE_PATH + '/ai-turn-end.txt',
            hand=str(player.hand)
        )

    @staticmethod
    def discarded(discard):
        return TextTemplate.render(
            TEMPLATE_PATH + '/player-discarded.txt',
            discard=discard
        )

    @staticmethod
    def discard_data(current_score, scores):
        return TextTemplate.render(
            TEMPLATE_PATH + '/ai-discard-data.txt',
            current_score=str(current_score),
            scores=str(scores),
            min_score=str(min(scores))
        )

    @staticmethod
    def hand_data(score):
        return TextTemplate.render(
            TEMPLATE_PATH + '/ai-hand-data.txt',
            score=str(score)
        )
